import nltk
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import warnings

# Font Initialization with System-Wide Check
def initialize_fonts():
    """
    Initializes and sets the fonts for matplotlib, checking for system-installed fonts.
    """
    font_issue = False  # Flag to track if there are any issues with fonts

    # Emoji-compatible fonts to check
    emoji_fonts = ["Segoe UI Emoji", "Apple Color Emoji", "Noto Emoji"]

    # System-wide font list
    available_fonts = {f.name for f in fm.fontManager.ttflist}

    # Find available emoji-compatible font
    selected_emoji_font = None
    for font in emoji_fonts:
        if font in available_fonts:
            selected_emoji_font = font
            break

    # Set font families
    if selected_emoji_font:
        plt.rcParams["font.family"] = [selected_emoji_font, "sans-serif"]
    else:
        warnings.warn(
            "No emoji-compatible font found. Using default 'sans-serif'. "
            "Install 'Segoe UI Emoji', 'Apple Color Emoji', or 'Noto Emoji' for better emoji support."
        )
        plt.rcParams["font.family"] = ["sans-serif"]

    # Check for system-wide Roboto font
    if "Roboto" in available_fonts:
        plt.rcParams["font.family"].insert(0, "Roboto")

initialize_fonts()

def ensure_nltk_resources(resources):
    """
    Ensure that the required NLTK resources are available.
    Downloads the resources only if they are not already present.

    Args:
        resources (list of tuples): List of resources to check, each as a tuple
            (resource_name, resource_type).
    """
    for resource_name, _ in resources:
            nltk.download(resource_name)

# List of required NLTK resources
required_resources = [
    ('punkt', 'tokenizers'),
    ('stopwords', 'corpora'),
    ('vader_lexicon', 'sentiment'),
    ('averaged_perceptron_tagger', 'taggers'),
    ('wordnet', 'corpora'),
]

# Ensure resources are available
ensure_nltk_resources(required_resources)



# analyzer.py (inside whatsapp_analyzer)
import os
from .parser import Parser
from .utils import (
    df_basic_cleanup,
)
# analysis_utils contains basic_stats, which in turn uses plot_utils and other analysis functions.
# analyzer.py only needs to import basic_stats from analysis_utils.
from .analysis_utils import basic_stats
from .constants import html_template # Used directly in generate_report

# Unused imports that were previously kept "for now" or for indirect dependencies:
# from .plot_utils import (
#     clean_message,
#     extract_emojis,
#     apply_consistent_plot_styling,
#     plot_activity_heatmap,
#     plot_sentiment_distribution,
#     plot_most_active_hours,
#     generate_wordcloud,
#     analyze_language_complexity,
#     plot_response_time_distribution,
#     analyze_sentiment_over_time,
#     analyze_emotion_over_time,
#     plot_emoji_usage,
#     plot_sentiment_bubble,
#     plot_vocabulary_diversity,
#     plot_language_complexity_pos,
#     plot_user_relationship_graph,
#     plot_skills_radar_chart,
# )
# from .analysis_utils import (
#     analyze_behavioral_traits, # Used by basic_stats
#     generate_behavioral_insights_text, # Used by basic_stats
#     analyze_hindi_abuse, # Used by basic_stats
# )
# from .constants import (
#     custom_hinglish_stopwords, # Referenced in analysis_utils via stop_words
#     skill_keywords, # Referenced in analysis_utils
#     hindi_abusive_words, # Referenced in analysis_utils
#     stop_words # Referenced in analysis_utils
# )
# External library imports that are no longer directly used in this file:
# import warnings
# import pandas as pd
# from textblob import TextBlob
# import nltk
# from nltk.corpus import stopwords
# from collections import Counter
# import re
# from sklearn.feature_extraction.text import CountVectorizer
# from wordcloud import WordCloud
# import emoji
# import base64
# from io import BytesIO
# from functools import lru_cache
# import networkx as nx
# import matplotlib.font_manager as fm
# import matplotlib.pyplot as plt

class WhatsAppAnalyzer:
    def __init__(self, chat_file, out_dir="."):
        self.chat_file = chat_file
        self.out_dir = out_dir
        self.parser = Parser(self.chat_file)
        self.df = df_basic_cleanup(self.parser.parse_chat_data())

    def generate_report(self, users=None):
        """
        Generates HTML reports for specified users.

        Args:
            users (list, optional): A list of usernames for which to generate reports. 
                                   If None, reports are generated for all users. Defaults to None.
        """
        if users is None:
            users = self.df["name"].unique()

        for name in users:
            if name != "System":
                # Call basic_stats from analysis_utils
                user_stats = basic_stats(self.df, name) 
                
                # Generate HTML for top 5 emojis and add to user_stats
                # user_stats["Top 5 Emojis"] is a list of (emoji, count) tuples
                top_5_emojis_html_str = " ".join(
                    [f"{emoji_val} ({count})" for emoji_val, count in user_stats["Top 5 Emojis"]]
                )
                user_stats['top_5_emojis_html'] = top_5_emojis_html_str
                
                # The n-gram and abuse count formatting is now done in basic_stats.
                # html_template will directly use these pre-formatted HTML strings.
                # We can now use dictionary unpacking for formatting.
                final_html = html_template.format(name=name, **user_stats)

                output_path = os.path.join(
                    self.out_dir, f"{name}_report.html"
                )
                os.makedirs(os.path.dirname(output_path), exist_ok=True)

                with open(output_path, "w", encoding="utf-8") as output_file:
                    output_file.write(final_html)

                print(f"Report for {name} has been generated and saved at {output_path}")
    
# Example usage (you can put this in a separate script or in your main function):
# if __name__ == "__main__":
#     analyzer = WhatsAppAnalyzer(chat_file="../data/whatsapp_chat.txt", out_dir="../data")
#     analyzer.generate_report()


import re
from dateutil import parser
import pandas as pd

class Parser:
    def __init__(self, file_path):
        """
        Initializes the Parser with the path to the WhatsApp chat file.

        Args:
            file_path (str): Path to the chat file.
        """
        self.file_path = file_path

    def preprocess_lines(self, lines):
        """
        Preprocesses the chat lines to remove line breaks if the next line
        does not start with a valid date pattern.

        Args:
            lines (list): List of chat lines.

        Returns:
            list: Preprocessed chat lines.
        """
        processed_lines = []
        buffer = ""

        # Updated date_pattern
        date_pattern = re.compile(r"^\d{2}/\d{2}/\d{2,4}, \d{1,2}:\d{2}\s(?:AM|PM|am|pm)")

        for line in lines:
            if date_pattern.match(line):
                if buffer:
                    processed_lines.append(buffer)
                buffer = line.strip()
            else:
                buffer += f" {line.strip()}"

        if buffer:  # Add the last buffered line
            processed_lines.append(buffer)

        return processed_lines

    def parse_chat_data(self):
        """Parses the WhatsApp chat file and returns a DataFrame."""
        with open(self.file_path, "r", encoding="utf-8") as file:
            chat_lines = file.readlines()

        # Preprocess lines to handle multiline messages
        chat_lines = self.preprocess_lines(chat_lines)

        chat_data = []
        for line in chat_lines:
            try:
                # Updated user message regex
                match = re.match(
                    r"(\d{2}/\d{2}/\d{2,4}, \d{1,2}:\d{2}\s(?:AM|PM|am|pm)) - (.*?): (.*)", line
                )
                if match:
                    timestamp, sender, message = match.groups()
                    date_obj = parser.parse(timestamp)
                    chat_data.append({"t": date_obj, "name": sender, "message": message})
                else:
                    # Updated system message regex
                    match2 = re.match(
                        r"(\d{2}/\d{2}/\d{2,4}, \d{1,2}:\d{2}\s(?:AM|PM|am|pm)) - (.*)", line
                    )
                    if match2:
                        timestamp, event = match2.groups()
                        date_obj = parser.parse(timestamp)
                        chat_data.append({"t": date_obj, "name": "System", "message": event})
                    else:
                        print(f"Skipping line: {line.strip()} (No match)")
            except (ValueError, AttributeError):
                print(f"Skipping line: {line.strip()} (Parse error)")

        return pd.DataFrame(chat_data)



# tests/test_utils.py

import unittest
import pandas as pd
import numpy as np
from datetime import datetime, date, time
import os
import tempfile

from whatsapp_analyzer.utils import df_basic_cleanup, anonymize, ANIMAL_NAMES
from whatsapp_analyzer.parser import Parser

class TestUtils(unittest.TestCase):

    def setUp(self):
        # Create a temporary file path for chat data
        self.temp_file_handle, self.temp_file_path = tempfile.mkstemp(suffix=".txt", text=True)
        os.close(self.temp_file_handle) # Close the handle, we'll open/write/close as needed in tests

    def tearDown(self):
        # Clean up the temporary file
        if os.path.exists(self.temp_file_path):
            os.remove(self.temp_file_path)

    def _create_chat_file(self, content_lines):
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            for line in content_lines:
                f.write(line + "\n") # Ensure each entry is a line
        return self.temp_file_path

    def test_df_basic_cleanup_columns_exist(self):
        """Test that all expected columns are created by df_basic_cleanup."""
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Hello world",
            "10/11/22, 11:30 AM - User2: <Media omitted>"
        ]
        chat_file = self._create_chat_file(chat_lines)
        
        parser = Parser(chat_file)
        raw_df = parser.parse_chat_data()
        
        # Ensure raw_df is not empty and has 't' column before cleanup
        self.assertFalse(raw_df.empty, "Raw DataFrame should not be empty")
        self.assertIn('t', raw_df.columns, "'t' column must be in raw_df for cleanup")

        clean_df = df_basic_cleanup(raw_df.copy())

        expected_columns = [
            'date_time', 'date', 'year', 'month', 'monthn', 'day', 'dayn', 
            'woy', 'doy', 'dow', 'ym', 'yw', 'yd', 'md', 'time', 'hour', 
            'min', 'hm', 'name', 'message', 'mlen', 'emoji', 'emojicount', 
            'urls', 'urlcount', 'yturls', 'yturlcount', 'mediacount', 
            'editcount', 'deletecount'
        ]
        
        missing_columns = [col for col in expected_columns if col not in clean_df.columns]
        self.assertEqual(len(missing_columns), 0, f"Missing columns: {missing_columns}")
        
        extra_columns = [col for col in clean_df.columns if col not in expected_columns]
        self.assertEqual(len(extra_columns), 0, f"Extra columns found: {extra_columns}")


    def test_df_basic_cleanup_data_types(self):
        """Test data types of key columns after df_basic_cleanup."""
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Test message http://example.com 😊",
            "10/11/22, 11:30 AM - User2: <Media omitted>"
        ]
        chat_file = self._create_chat_file(chat_lines)
        parser = Parser(chat_file)
        raw_df = parser.parse_chat_data()
        clean_df = df_basic_cleanup(raw_df.copy())

        self.assertFalse(clean_df.empty, "Cleaned DataFrame should not be empty for type checking")

        self.assertTrue(pd.api.types.is_datetime64_any_dtype(clean_df['date_time']), "date_time column should be datetime type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['emojicount']), "emojicount column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['urlcount']), "urlcount column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['hour']), "hour column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['mlen']), "mlen column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['mediacount']), "mediacount column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['editcount']), "editcount column should be integer type")
        self.assertTrue(pd.api.types.is_integer_dtype(clean_df['deletecount']), "deletecount column should be integer type")
        self.assertEqual(clean_df['date'].dtype, np.dtype('object')) # pandas stores date objects as 'object'
        self.assertTrue(isinstance(clean_df['date'].iloc[0], date), "date column elements should be datetime.date")


    def test_df_basic_cleanup_calculated_values(self):
        """Test specific calculated values from df_basic_cleanup."""
        # Note: 10/10/2022 was a Monday
        chat_lines = [
            "10/10/22, 10:05 AM - User1: Hello world http://example.com 😊",
            "10/11/22, 11:15 AM - User2: <Media omitted>",
            "10/12/22, 12:20 PM - User1: This message was edited", # This is not the WhatsApp system message for edits
            "10/13/22, 01:25 PM - User1: <This message was edited>", # This IS the system message
            "10/14/22, 02:30 PM - User2: This message was deleted", # This IS a system message for deletion
            "10/15/22, 03:35 PM - User1: You deleted this message" # This IS also a system message for deletion
        ]
        chat_file = self._create_chat_file(chat_lines)
        parser = Parser(chat_file)
        raw_df = parser.parse_chat_data()
        clean_df = df_basic_cleanup(raw_df.copy())

        self.assertEqual(len(clean_df), 6)

        # Test first message (User1)
        msg1 = clean_df.iloc[0]
        self.assertEqual(msg1['name'], "User1")
        self.assertEqual(msg1['message'], "Hello world http://example.com 😊")
        self.assertEqual(msg1['mlen'], len("Hello world http://example.com 😊"))
        self.assertEqual(msg1['emojicount'], 1)
        self.assertEqual(msg1['urlcount'], 1)
        self.assertEqual(msg1['mediacount'], 0)
        self.assertEqual(msg1['editcount'], 0)
        self.assertEqual(msg1['deletecount'], 0)
        self.assertEqual(msg1['hour'], 10)
        self.assertEqual(msg1['dayn'], 'Monday')
        self.assertEqual(msg1['monthn'], 'October')
        self.assertEqual(msg1['year'], 2022)
        self.assertEqual(msg1['hm'], 10.08) # 10 + 5/60

        # Test second message (User2) - Media
        msg2 = clean_df.iloc[1]
        self.assertEqual(msg2['name'], "User2")
        self.assertEqual(msg2['message'], "<Media omitted>")
        self.assertEqual(msg2['mediacount'], 1)

        # Test third message (User1) - Contains "This message was edited" but NOT the system way
        msg3 = clean_df.iloc[2]
        self.assertEqual(msg3['editcount'], 0) 

        # Test fourth message (User1) - Actual edited message
        msg4 = clean_df.iloc[3]
        self.assertEqual(msg4['editcount'], 1)

        # Test fifth message (User2) - "This message was deleted"
        msg5 = clean_df.iloc[4]
        self.assertEqual(msg5['deletecount'], 1)
        
        # Test sixth message (User1) - "You deleted this message"
        msg6 = clean_df.iloc[5]
        self.assertEqual(msg6['deletecount'], 1)

# --- TestAnonymize class starts here ---
class TestAnonymize(unittest.TestCase):

    def setUp(self):
        # Create temporary files for input and output
        self.temp_input_file = tempfile.NamedTemporaryFile(mode="w", delete=False, encoding="utf-8")
        self.temp_output_file = tempfile.NamedTemporaryFile(mode="r", delete=False, encoding="utf-8")
        
        # Store their paths
        self.input_path = self.temp_input_file.name
        self.output_path = self.temp_output_file.name
        
        # Close them immediately, anonymize function will open/close them
        self.temp_input_file.close()
        self.temp_output_file.close()

    def tearDown(self):
        # Clean up the temporary files
        os.remove(self.input_path)
        os.remove(self.output_path)

    def _write_to_input(self, lines):
        with open(self.input_path, 'w', encoding='utf-8') as f:
            for line in lines:
                f.write(line + '\n')

    def _read_file_lines(self, file_path):
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                return [line.strip() for line in f.readlines()]
        except FileNotFoundError:
            return [] # Return empty list if file not found (e.g. for not_found test)

    def test_basic_anonymization(self):
        input_lines = [
            "20/03/2023, 10:00 - Alice: Hello Bob",
            "20/03/2023, 10:01 - Bob: Hi Alice",
            "20/03/2023, 10:02 - Alice: How are you?"
        ]
        self._write_to_input(input_lines)
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)
        
        animal1 = ANIMAL_NAMES[0]
        animal2 = ANIMAL_NAMES[1]
        
        expected_lines = [
            f"20/03/2023, 10:00 - user_1_{animal1}: Hello Bob",
            f"20/03/2023, 10:01 - user_2_{animal2}: Hi Alice",
            f"20/03/2023, 10:02 - user_1_{animal1}: How are you?"
        ]
        self.assertEqual(output_lines, expected_lines)

    def test_system_messages(self):
        input_lines = [
            "20/03/2023, 10:00 - System: Alice added Bob",
            "20/03/2023, 10:01 - Carol: Hello everyone",
            "This is a line that won't be parsed by _parse_line."
        ]
        self._write_to_input(input_lines)
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)
        
        animal1 = ANIMAL_NAMES[0] # Carol will be user_1

        expected_lines = [
            "20/03/2023, 10:00 - System: Alice added Bob",
            f"20/03/2023, 10:01 - user_1_{animal1}: Hello everyone",
            "This is a line that won't be parsed by _parse_line."
        ]
        self.assertEqual(output_lines, expected_lines)

    def test_username_in_message_content(self):
        input_lines = [
            "20/03/2023, 10:00 - Dave: Hey Eve, how is Dave doing?",
            "20/03/2023, 10:01 - Eve: Dave is fine."
        ]
        self._write_to_input(input_lines)
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)

        animal1 = ANIMAL_NAMES[0] # Dave
        animal2 = ANIMAL_NAMES[1] # Eve
        
        expected_lines = [
            f"20/03/2023, 10:00 - user_1_{animal1}: Hey Eve, how is Dave doing?",
            f"20/03/2023, 10:01 - user_2_{animal2}: Dave is fine."
        ]
        self.assertEqual(output_lines, expected_lines)

    def test_empty_input_file(self):
        self._write_to_input([])
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)
        self.assertEqual(output_lines, [])

    def test_no_matching_chat_lines(self):
        input_lines = [
            "This is just some random text.",
            "Another line without WhatsApp format.",
            "12345"
        ]
        self._write_to_input(input_lines)
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)
        # Expect output to be same as input, plus newline handling (strip ensures comparison is fair)
        self.assertEqual(output_lines, input_lines)

    def test_animal_list_cycling(self):
        num_users = len(ANIMAL_NAMES) + 2  # e.g., 12 users if 10 animals
        input_lines = []
        expected_lines = []
        for i in range(num_users):
            user_name = f"User{i+1}"
            input_lines.append(f"20/03/2023, 10:{i:02d} - {user_name}: Message {i+1}")
            
            animal_index = i % len(ANIMAL_NAMES)
            expected_animal = ANIMAL_NAMES[animal_index]
            expected_lines.append(f"20/03/2023, 10:{i:02d} - user_{i+1}_{expected_animal}: Message {i+1}")
            
        self._write_to_input(input_lines)
        anonymize(self.input_path, self.output_path)
        output_lines = self._read_file_lines(self.output_path)
        self.assertEqual(output_lines, expected_lines)

    def test_input_file_not_found(self):
        # Ensure the input file does not exist
        non_existent_input_path = "/tmp/this_file_should_definitely_not_exist_for_test.txt"
        if os.path.exists(non_existent_input_path):
             os.remove(non_existent_input_path) # Just in case

        # Suppress print output from anonymize for this test
        import sys
        from io import StringIO
        original_stdout = sys.stdout
        sys.stdout = captured_output = StringIO()

        anonymize(non_existent_input_path, self.output_path)
        
        sys.stdout = original_stdout # Restore stdout
        
        # Check that an error message was printed (optional, but good to confirm)
        # self.assertIn("Error: Input file", captured_output.getvalue())

        # Check that the output file was not created or is empty
        output_lines = self._read_file_lines(self.output_path)
        self.assertEqual(output_lines, [], "Output file should be empty or not created if input is not found.")


if __name__ == '__main__':
    unittest.main()



# whatsapp_analyzer/plot_utils.py
import base64
import emoji
import re
from io import BytesIO
from collections import Counter
from functools import lru_cache

import matplotlib.pyplot as plt
import matplotlib.font_manager as fm
import networkx as nx
import nltk
import pandas as pd
import seaborn as sns
from sklearn.feature_extraction.text import CountVectorizer
from textblob import TextBlob
from wordcloud import WordCloud

from .constants import custom_hinglish_stopwords, skill_keywords # Import from constants within the package

# Combine NLTK stopwords with custom Hinglish stopwords
stop_words = set(nltk.corpus.stopwords.words('english')).union(custom_hinglish_stopwords)

@lru_cache(maxsize=None)  # Cache all unique calls
def clean_message(msg):
    """
    Clean the message by removing URLs, media omitted phrases, and trimming spaces.
    """
    # Remove URLs
    msg = re.sub(r'http[s]?://\S+', '', msg)
    # Remove "media omitted" phrases, case-insensitive
    msg = re.sub(r'\b(media omitted|<media omitted>)\b', '', msg, flags=re.IGNORECASE)
    # Strip any extra spaces
    msg = msg.strip()
    return msg

def extract_emojis(text):
    """Extract emojis from text."""
    return [c for c in text if c in emoji.EMOJI_DATA]

def plot_to_base64(plt):
    """Convert a Matplotlib plot to a base64 encoded image."""
    img_buffer = BytesIO()
    plt.savefig(img_buffer, format='png', bbox_inches='tight')
    img_buffer.seek(0)
    img_base64 = base64.b64encode(img_buffer.getvalue()).decode('utf-8')
    plt.close()
    return img_base64

def apply_consistent_plot_styling(plt, title, xlabel, ylabel):
    """Applies consistent styling to Matplotlib plots."""
    plt.title(title, fontsize=14)
    plt.xlabel(xlabel, fontsize=12)
    plt.ylabel(ylabel, fontsize=12)
    plt.xticks(fontsize=10)
    plt.yticks(fontsize=10)
    #plt.tight_layout()

def plot_activity_heatmap(df, username=None):
    """Plot an activity heatmap and return base64 image."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['date'] = pd.to_datetime(df_filtered['date'])
    df_filtered['weekday'] = df_filtered['date'].dt.day_name()
    df_filtered['hour'] = df_filtered['hour']

    heatmap_data = df_filtered.pivot_table(index='weekday', columns='hour', values='message', aggfunc='count', fill_value=0)
    heatmap_data = heatmap_data.reindex(['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'])

    plt.figure(figsize=(12, 6), constrained_layout=True)
    sns.heatmap(heatmap_data, cmap='viridis', annot=False, cbar_kws={'label': 'Number of Messages'})
    apply_consistent_plot_styling(plt, f'Activity Heatmap {"for " + username if username else ""}', 'Hour of the Day', 'Day of the Week')
    return plot_to_base64(plt)

def plot_sentiment_distribution(df, username=None):
    """Plot sentiment distribution and return base64 image."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['sentiment'] = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)

    plt.figure(figsize=(8, 5))
    sns.histplot(df_filtered['sentiment'], bins=20, kde=True, color='skyblue')
    apply_consistent_plot_styling(plt, f'Sentiment Distribution {"for " + username if username else ""}', 'Sentiment Polarity', 'Frequency')
    return plot_to_base64(plt)

def plot_most_active_hours(df, username=None):
    """Plot a bar chart of the most active hours and return base64 image."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    message_counts_by_hour = df_filtered['hour'].value_counts().sort_index()

    plt.figure(figsize=(12, 6), constrained_layout=True)
    plt.bar(message_counts_by_hour.index, message_counts_by_hour.values, color='skyblue')
    apply_consistent_plot_styling(plt, f'Most Active Hours {"for " + username if username else ""}', 'Hour of the Day', 'Number of Messages')
    return plot_to_base64(plt)

def generate_wordcloud(df, username=None):
    """Generate word cloud and return base64 image."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['clean_message'] = df_filtered['message'].apply(lambda x: clean_message(str(x)))
    text = " ".join(msg for msg in df_filtered['clean_message'] if isinstance(msg, str) and len(msg.strip())>0)

    plt.figure(figsize=(10, 8))
    if not text: # Handle case with no text for word cloud
        plt.text(0.5, 0.5, "No words to display in word cloud.", ha='center', va='center', fontsize=12)
    else:
        try:
            wordcloud = WordCloud(stopwords=stop_words, background_color="white").generate(text)
            plt.imshow(wordcloud, interpolation='bilinear')
        except ValueError as e: # Catch any other potential errors from WordCloud
            plt.text(0.5, 0.5, f"Could not generate word cloud:\n{e}", ha='center', va='center', fontsize=12, color='red')
            
    plt.axis("off")
    plt.title(f'Word Cloud {"for " + username if username else ""}', fontsize=14)
    #plt.tight_layout()
    return plot_to_base64(plt)

def analyze_language_complexity(df, username=None):
    """Analyze language complexity and return base64 images."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['clean_message'] = df_filtered['message'].apply(lambda x: clean_message(str(x)))
    
    # filter out emojis
    df_filtered['word_length'] = df_filtered['clean_message'].apply(
        lambda x: [len(word) for word in str(x).split() if word.lower() not in stop_words and len(word) > 1 and not all(c in emoji.EMOJI_DATA for c in word)]
    )
    
    avg_word_lengths = df_filtered['word_length'].apply(lambda x: sum(x) / len(x) if len(x) > 0 else 0)

    # Handle cases with only emojis or empty messages
    df_filtered['sentence_length'] = df_filtered['clean_message'].apply(
        lambda x: len(nltk.sent_tokenize(str(x))) if str(x).strip() else 0
    )
    avg_sentence_lengths = df_filtered['sentence_length'].apply(
        lambda x: len(str(x).split()) / x if x > 0 and len(str(x).split()) > 0 else 0
    )

    fig, axs = plt.subplots(1, 2, figsize=(12, 5))

    sns.histplot(avg_word_lengths, bins=20, kde=True, color='skyblue', ax=axs[0])
    apply_consistent_plot_styling(plt, f'Average Word Length {"for " + username if username else ""}', 'Average Word Length', 'Frequency')

    sns.histplot(avg_sentence_lengths, bins=20, kde=True, color='salmon', ax=axs[1])
    apply_consistent_plot_styling(plt, f'Average Sentence Length {"for " + username if username else ""}', 'Average Sentence Length (words)', 'Frequency')

    #plt.tight_layout()
    
    # Convert the combined plot to base64
    combined_plot_base64 = plot_to_base64(plt)
    
    return combined_plot_base64

def plot_response_time_distribution(response_times, username=None):
    """Plot the distribution of response times."""
    plt.figure(figsize=(8, 5))
    sns.histplot(response_times, bins=20, kde=True, color='skyblue')
    apply_consistent_plot_styling(plt, f'Response Time Distribution {"for " + username if username else ""}', 'Response Time (minutes)', 'Frequency')
    return plot_to_base64(plt)

def analyze_sentiment_over_time(df, username=None):
    """Analyze sentiment over time and return base64 image of the plot."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['sentiment'] = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)
    df_filtered['date'] = pd.to_datetime(df_filtered['date'])
    df_filtered.set_index('date', inplace=True)

    # Resample to daily frequency and calculate the mean sentiment
    daily_sentiment = df_filtered['sentiment'].resample('W').mean()

    plt.figure(figsize=(12, 6))
    plt.plot(daily_sentiment.index, daily_sentiment.values, color='purple')
    apply_consistent_plot_styling(plt, f'Sentiment Over Time {"for " + username if username else ""}', 'Date', 'Average Sentiment')
    
    return plot_to_base64(plt)

def analyze_emotion_over_time(df, username=None):
    """Analyze emotion over time using TextBlob and return base64 image of the plot."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()
    
    df_filtered['date'] = pd.to_datetime(df_filtered['date'])
    df_filtered.set_index('date', inplace=True)

    # Define a function to categorize sentiment into emotions
    def categorize_emotion(score):
        if score > 0.5:
            return "joy"
        elif score > 0:
            return "surprise"
        elif score < -0.5:
            return "sadness"
        elif score < 0:
            return "anger"
        else:
            return "neutral"
    
    # Apply sentiment analysis and emotion categorization
    df_filtered['sentiment'] = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)
    df_filtered['emotion'] = df_filtered['sentiment'].apply(categorize_emotion)
    
    # Resample to daily frequency and count the occurrences of each emotion
    daily_emotions = df_filtered.groupby(pd.Grouper(freq='D'))['emotion'].apply(lambda x: x.value_counts()).unstack(fill_value=0)
    
    
    plt.figure(figsize=(12, 6))
    for emotion in daily_emotions.columns:
        plt.plot(daily_emotions.index, daily_emotions[emotion], label=emotion)
    plt.legend()
    apply_consistent_plot_styling(plt, f'Emotion Trends Over Time {"for " + username if username else ""}', 'Date', 'Emotion Score')
    
    return plot_to_base64(plt)

def plot_emoji_usage(df, username=None):
    """Plot a bar chart of the top 5 emojis used and return base64 image."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    df_filtered['emojis'] = df_filtered['message'].apply(extract_emojis)
    all_emojis = [emoji for sublist in df_filtered['emojis'] for emoji in sublist]
    top_emojis = Counter(all_emojis).most_common(5)

    if not top_emojis: # Handle case with no emojis
        plt.figure(figsize=(10, 6))
        plt.text(0.5, 0.5, "No emojis found.", ha='center', va='center', fontsize=12)
        apply_consistent_plot_styling(plt, f'Emoji Usage {"for " + username if username else ""}', 'Emoji', 'Count')
    else:
        emojis, counts = zip(*top_emojis)
        plt.figure(figsize=(10, 6))
        plt.bar(emojis, counts, color='skyblue')
        apply_consistent_plot_styling(plt, f'Emoji Usage {"for " + username if username else ""}', 'Emoji', 'Count')
    return plot_to_base64(plt)

def plot_sentiment_bubble(df, username=None):
    """
    Plot a bubble chart of sentiment distribution and return base64 image.
    x-axis: Polarity (Positive/Negative)
    y-axis: Subjectivity (Objective/Subjective)
    Bubble size: Number of messages
    """
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    # Calculate sentiment polarity and subjectivity
    df_filtered['polarity'] = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)
    df_filtered['subjectivity'] = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.subjectivity)

    # Count the number of messages for each sentiment
    sentiment_counts = df_filtered.groupby(['polarity', 'subjectivity']).size().reset_index(name='counts')

    plt.figure(figsize=(10, 8))
    plt.scatter(sentiment_counts['polarity'], sentiment_counts['subjectivity'], s=sentiment_counts['counts']*10, alpha=0.6, color='purple')
    apply_consistent_plot_styling(plt, f'Sentiment Distribution {"for " + username if username else ""}', 'Polarity (Positive/Negative)', 'Subjectivity (Objective/Subjective)')
    return plot_to_base64(plt)

def plot_vocabulary_diversity(df, username=None):
    """
    Plot a scatter plot of vocabulary diversity over time and return base64 image.
    x-axis: Unique words used
    y-axis: Average message length
    """
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()
        
    df_filtered['clean_message_lower'] = df_filtered['clean_message'].str.lower()
    corpus = df_filtered['clean_message_lower'].dropna()
    
    unique_words_count = 0
    if not corpus.empty:
        vectorizer = CountVectorizer(stop_words=list(stop_words))
        try:
            word_matrix = vectorizer.fit_transform(corpus)
            unique_words_count = len(vectorizer.get_feature_names_out())
        except ValueError: # Handles empty vocabulary
            unique_words_count = 0

    # Calculate average message length
    avg_msg_len_val = df_filtered['message'].apply(lambda x: len(str(x).split())).mean()
    if pd.isna(avg_msg_len_val): # Handle case where mean might be NaN (e.g., no messages)
        avg_msg_len_val = 0


    plt.figure(figsize=(10, 8))
    if unique_words_count > 0:
        plt.scatter(unique_words_count, avg_msg_len_val, color='green')
    else:
        plt.text(0.5, 0.5, "Not enough data for vocabulary diversity.", ha='center', va='center', fontsize=12)
        # Still set x/y limits to make the plot consistent if needed, or adjust as per visualization strategy
        plt.xlim(0, 1) # Example limits
        plt.ylim(0, 1) # Example limits

    apply_consistent_plot_styling(plt, f'Vocabulary Diversity {"for " + username if username else ""}', 'Unique Words (Count)', 'Average Message Length (Words)')
    return plot_to_base64(plt)

def plot_language_complexity_pos(df, username=None):
    """
    Analyze and plot the distribution of POS tags for a user or the entire chat,
    and return a base64 image of the plot.
    """
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    # Function to extract POS tags from a message
    def extract_pos_tags(message):
        analysis = TextBlob(message)
        return [tag for (word, tag) in analysis.tags]

    # Apply POS tag extraction to each message
    df_filtered['pos_tags'] = df_filtered['message'].apply(extract_pos_tags)

    # Flatten the list of POS tags and count their occurrences
    all_pos_tags = [tag for sublist in df_filtered['pos_tags'] for tag in sublist]
    pos_counts = Counter(all_pos_tags)

    # Convert to DataFrame for plotting
    pos_df = pd.DataFrame(list(pos_counts.items()), columns=['POS Tag', 'Count'])

    # Plotting
    plt.figure(figsize=(12, 6))
    sns.barplot(x='POS Tag', y='Count', data=pos_df, color='skyblue')
    apply_consistent_plot_styling(plt, f'POS Tag Distribution {"for " + username if username else ""}', 'POS Tag', 'Count')

    return plot_to_base64(plt)

def plot_user_relationship_graph(df):
    """
    Plot a graph representing the relationships between users based on message interactions.
    Nodes represent users, and edges represent interactions between them.
    """
    df = df[df['name'] != "System"].reset_index(drop=True)
    
    # Create a graph
    G = nx.Graph()

    # Add nodes for each user
    for user in df['name'].unique():
        G.add_node(user)

    # Analyze interactions and add edges
    for i in range(len(df) - 1):
        sender = df['name'].iloc[i]
        next_sender = df['name'].iloc[i + 1]
        if sender != next_sender:
            if G.has_edge(sender, next_sender):
                G[sender][next_sender]['weight'] += 1
            else:
                G.add_edge(sender, next_sender, weight=1)

    # Draw the graph
    plt.figure(figsize=(12, 10))
    pos = nx.spring_layout(G, k=0.5, iterations=50)
    nx.draw(G, pos, with_labels=True, node_size=2000, node_color='skyblue', font_size=8, width=[d['weight'] / 5 for u, v, d in G.edges(data=True)])
    edge_labels = nx.get_edge_attributes(G, 'weight')
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_size=8)
    plt.title("User Relationship Graph", fontsize=15)
    #plt.tight_layout()

    return plot_to_base64(plt)

def plot_skills_radar_chart(df, username=None):
    """
    Generate a radar chart to visualize various skills based on keyword analysis.
    """
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    

    # Count keyword occurrences for each skill
    skill_counts = {}
    for skill, keywords in skill_keywords.items():
        skill_counts[skill] = sum(df_filtered['clean_message'].str.lower().str.count('|'.join(keywords)))

    # Prepare data for radar chart
    skills = list(skill_counts.keys())
    counts = list(skill_counts.values())
    
    # Normalize counts for radar chart
    if counts:
        max_val = max(counts)
        if max_val == 0: # All skill counts are zero
             # Prevent division by zero; keep normalized_counts as zeros or handle as appropriate
            normalized_counts = [0.0] * len(counts)
        else:
            normalized_counts = [c / max_val for c in counts]
    else: # No skills defined or counts list is empty for some reason
        normalized_counts = []
        skills = [] # Ensure skills is also empty if counts is empty

    # Number of variables (skills)
    num_vars = len(skills)

    # Compute angle for each axis
    angles = [n / float(num_vars) * 2 * 3.14159 for n in range(num_vars)]
    angles += angles[:1]  # Complete the loop

    # Initialize radar chart
    plt.figure(figsize=(6, 6))
    ax = plt.subplot(111, polar=True)

    # Draw one axis per variable and add labels
    plt.xticks(angles[:-1], skills, color='black', size=10)

    # Draw ylabels (normalized counts)
    ax.set_rlabel_position(0)
    plt.yticks([0.25, 0.5, 0.75, 1], ["0.25", "0.5", "0.75", "1"], color="grey", size=8)
    plt.ylim(0, 1)

    # Plot data
    ax.plot(angles, normalized_counts + normalized_counts[:1], linewidth=2, linestyle='solid')
    ax.fill(angles, normalized_counts + normalized_counts[:1], 'b', alpha=0.1)

    # Add title
    plt.title(f'Skills Radar Chart {"for " + username if username else ""}', size=15, color='black', y=1.1)
    #plt.tight_layout()

    return plot_to_base64(plt)



# whatsapp_analyzer/analysis_utils.py

import nltk
from textblob import TextBlob
from sklearn.feature_extraction.text import CountVectorizer
from collections import Counter
import pandas as pd # Though df is passed, good practice

# Imports from within the package
from .constants import (
    skill_keywords,
    hindi_abusive_words,
    stop_words # Assuming stop_words is comprehensive, including custom_hinglish_stopwords
)
from .plot_utils import (
    clean_message, # Used in basic_stats, analyze_behavioral_traits
    extract_emojis, # Used in basic_stats
    plot_activity_heatmap,
    plot_sentiment_distribution,
    plot_most_active_hours,
    generate_wordcloud,
    analyze_language_complexity,
    plot_response_time_distribution,
    analyze_sentiment_over_time,
    analyze_emotion_over_time,
    plot_emoji_usage,
    plot_sentiment_bubble,
    plot_vocabulary_diversity,
    plot_language_complexity_pos,
    plot_user_relationship_graph,
    plot_skills_radar_chart
)
# analyze_message_timing will be moved into this file.

# Function moved from utils.py
def analyze_message_timing(df, username=None):
    """Analyze the timing of messages and return response times."""
    if username:
        df_filtered = df[df['name'] == username].copy()
    else:
        df_filtered = df.copy()

    # Ensure 'date' column is datetime
    if not pd.api.types.is_datetime64_any_dtype(df_filtered['date']):
        df_filtered['date'] = pd.to_datetime(df_filtered['date'])

    # Use 'date_time' column for accurate time difference calculation
    df_filtered['time_diff'] = df_filtered.groupby('name')['date_time'].diff()
    response_times = df_filtered['time_diff'].dropna().apply(lambda x: x.total_seconds() / 60)  # in minutes

    return response_times

# Helper function to prepare user-specific data
def _prepare_user_data(df_orig, username=None):
    if username:
        df_filtered = df_orig[df_orig['name'] == username].copy()
    else:
        df_filtered = df_orig.copy()

    # Sentiment Analysis
    sentiments = df_filtered['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)
    positive_msgs = sum(sentiments > 0)
    negative_msgs = sum(sentiments < 0)

    # Time of Day Analysis
    def categorize_time_of_day(hour):
        if 6 <= hour < 12: return 'Morning'
        elif 12 <= hour < 16: return 'Mid-day'
        elif 16 <= hour < 18: return 'Evening'
        else: return 'Night'

    df_filtered['time_of_day'] = df_filtered['hour'].apply(categorize_time_of_day)
    morning_msgs = len(df_filtered[df_filtered['time_of_day'] == 'Morning'])
    midday_msgs = len(df_filtered[df_filtered['time_of_day'] == 'Mid-day'])
    evening_msgs = len(df_filtered[df_filtered['time_of_day'] == 'Evening'])
    night_msgs = len(df_filtered[df_filtered['time_of_day'] == 'Night'])
    message_counts_by_period = {'Morning': morning_msgs, 'Mid-day': midday_msgs, 'Evening': evening_msgs, 'Night': night_msgs}
    most_active_period = max(message_counts_by_period, key=message_counts_by_period.get) if message_counts_by_period else "N/A"

    # Clean messages - this is crucial for many subsequent analyses
    df_filtered['clean_message'] = df_filtered['message'].apply(lambda x: clean_message(str(x)))
    df_filtered['clean_message_lower'] = df_filtered['clean_message'].str.lower()
    
    return df_filtered, positive_msgs, negative_msgs, morning_msgs, midday_msgs, evening_msgs, night_msgs, most_active_period

# Helper function to calculate n-grams
def _calculate_ngrams(df_filtered):
    # Most Common n-grams
    def get_top_ngrams(corpus, n=1, top_k=10):
        # Ensure corpus is not empty and contains actual text
        if corpus.empty or corpus.str.strip().empty:
            return []
        try:
            vec = CountVectorizer(ngram_range=(n, n), stop_words=list(stop_words)).fit(corpus)
            bag_of_words = vec.transform(corpus)
            sum_words = bag_of_words.sum(axis=0)
            words_freq = [(word, sum_words[0, idx]) for word, idx in vec.vocabulary_.items()]
            # Filter for counts > 1
            words_freq_filtered = [item for item in words_freq if item[1] > 1]
            words_freq = sorted(words_freq_filtered, key=lambda x: x[1], reverse=True)
            return words_freq[:top_k]
        except ValueError: # Handles cases where vocabulary might be empty after stop word removal
            return []

    # Ensure there's data to process for n-grams
    corpus_for_ngrams = df_filtered['clean_message_lower'].dropna()
    common_unigrams = get_top_ngrams(corpus_for_ngrams, 1, 10)
    common_bigrams = get_top_ngrams(corpus_for_ngrams, 2, 10)
    common_trigrams = get_top_ngrams(corpus_for_ngrams, 3, 10)
    
    return common_unigrams, common_bigrams, common_trigrams

def analyze_behavioral_traits(df, username=None):
    """
    Analyze behavioral traits and return a dictionary of insights.
    Assumes df already has 'clean_message' and 'clean_message_lower' columns.
    """
    if username:
        # df_filtered is a copy, so modifications are safe.
        df_filtered_behavior = df[df['name'] == username].copy() 
    else:
        df_filtered_behavior = df.copy()

    # Ensure 'clean_message' column exists if not already present from _prepare_user_data
    if 'clean_message' not in df_filtered_behavior.columns:
        df_filtered_behavior['clean_message'] = df_filtered_behavior['message'].apply(lambda x: clean_message(str(x)))
    if 'clean_message_lower' not in df_filtered_behavior.columns:
        df_filtered_behavior['clean_message_lower'] = df_filtered_behavior['clean_message'].str.lower()

    traits = {}

    # --- Sentiment Analysis (already done in _prepare_user_data, but if called independently) ---
    if 'sentiment_polarity' not in df_filtered_behavior.columns:
        df_filtered_behavior['sentiment_polarity'] = df_filtered_behavior['message'].apply(lambda x: TextBlob(str(x)).sentiment.polarity)
    if 'sentiment_subjectivity' not in df_filtered_behavior.columns:
        df_filtered_behavior['sentiment_subjectivity'] = df_filtered_behavior['message'].apply(lambda x: TextBlob(str(x)).sentiment.subjectivity)
    
    traits['avg_sentiment_polarity'] = df_filtered_behavior['sentiment_polarity'].mean()
    traits['avg_sentiment_subjectivity'] = df_filtered_behavior['sentiment_subjectivity'].mean()


    # --- Psychometric Analysis ---
    traits['num_questions'] = df_filtered_behavior['message'].apply(lambda x: x.count('?')).sum()
    traits['num_exclamations'] = df_filtered_behavior['message'].apply(lambda x: x.count('!')).sum()
    traits['first_person_pronouns'] = df_filtered_behavior['clean_message'].str.lower().str.count(r'\b(i|me|my|mine|myself)\b').sum()

    # --- Skill Analysis (Keyword-based) ---
    traits['skills'] = {}
    for skill, keywords in skill_keywords.items():
        traits['skills'][skill] = sum(df_filtered_behavior['clean_message_lower'].str.count('|'.join(keywords)))


    # --- Language Complexity ---
    df_filtered_behavior['sentence_length_behavioral'] = df_filtered_behavior['clean_message'].apply(lambda x: len(nltk.sent_tokenize(str(x))) if str(x).strip() else 0)
    traits['avg_sentence_length'] = df_filtered_behavior['sentence_length_behavioral'].mean()


    # --- Lexical Diversity ---
    vectorizer = CountVectorizer(stop_words=list(stop_words))
    corpus_for_lexical = df_filtered_behavior['clean_message_lower'].dropna()
    unique_words_count = 0
    if not corpus_for_lexical.empty:
        try:
            word_matrix = vectorizer.fit_transform(corpus_for_lexical)
            unique_words_count = len(vectorizer.get_feature_names_out())
        except ValueError: 
            unique_words_count = 0 
        
    total_words_count = df_filtered_behavior['message'].apply(lambda x: len(str(x).split())).sum()
    traits['lexical_diversity'] = unique_words_count / total_words_count if total_words_count > 0 else 0

    return traits

def generate_behavioral_insights_text(traits, most_active_period, avg_response_time):
    """
    Generate human-readable insights based on behavioral traits.
    """
    insights = []

    # Sentiment Hints
    if traits.get('avg_sentiment_polarity', 0) > 0.2:
        insights.append("Tends to express positive sentiment in messages.")
    elif traits.get('avg_sentiment_polarity', 0) < -0.2:
        insights.append("Tends to express negative sentiment in messages.")
    else:
        insights.append("Maintains a neutral tone in messages.")

    if traits.get('avg_sentiment_subjectivity', 0) > 0.5:
        insights.append("Expresses subjective opinions and evaluations.")
    else:
        insights.append("Tends to communicate more objectively.")

    # Psychometric Hints
    if traits.get('num_questions', 0) > 20:
        insights.append("Asks a lot of questions, possibly indicating curiosity or a need for clarification.")
    if traits.get('num_exclamations', 0) > 5:
        insights.append("Uses exclamations frequently, suggesting excitement or strong opinions.")
    if traits.get('first_person_pronouns', 0) > 10:
        insights.append("Often refers to themselves, which might indicate a focus on personal experiences or opinions.")

    # Skill Hints
    user_skills = traits.get('skills', {})
    if user_skills.get('communication', 0) > 5:
        insights.append("Demonstrates strong communication skills based on keyword analysis.")
    if user_skills.get('technical', 0) > 5:
        insights.append("Exhibits technical skills based on keyword analysis.")
    if user_skills.get('leadership', 0) > 2:
        insights.append("Shows potential leadership qualities based on keyword analysis.")
    if user_skills.get('problem_solving', 0) > 5:
        insights.append("Appears to have good problem-solving skills based on keyword analysis.")
    if user_skills.get('teamwork', 0) > 5:
        insights.append("Likely a good team player based on keyword analysis.")

    # Timing Hints
    if avg_response_time is not None:
        if avg_response_time < 60: # Assuming minutes
            insights.append("Responds quickly to messages, indicating high engagement.")
        elif avg_response_time > 180: # Assuming minutes
            insights.append("Takes longer to respond, suggesting lower engagement or a busy schedule.")
        else:
            insights.append("Has a moderate response time.")

    if most_active_period is not None:
        if most_active_period == 'Morning':
            insights.append("Most active in the morning.")
        elif most_active_period == 'Mid-day':
            insights.append("Most active in the afternoon.")
        elif most_active_period == 'Evening':
            insights.append("Most active in the evening.")
        else: # Night
            insights.append("Most active at night.")
    
    # Language Complexity Hints
    if traits.get('avg_sentence_length', 0) > 3: # Assuming this is words per sentence from analyze_behavioral_traits
        insights.append("Uses long and complex sentences.")
    else:
        insights.append("Uses short and concise sentences.")

    # Lexical Diversity Hints
    if traits.get('lexical_diversity', 0) > 0.7:
        insights.append("Exhibits high lexical diversity, indicating a broad vocabulary.")
    elif traits.get('lexical_diversity', 0) < 0.4:
        insights.append("Has low lexical diversity, suggesting a more repetitive or focused communication style.")
    else:
        insights.append("Shows moderate lexical diversity.")

    return "<br/>".join(insights)

def analyze_hindi_abuse(df, username=None):
    """
    Analyze the use of Hindi abusive words and return a dictionary of counts.
    Assumes df has 'clean_message'.
    """
    if username:
        df_filtered_abuse = df[df['name'] == username].copy()
    else:
        df_filtered_abuse = df.copy()
    
    # Ensure 'clean_message' column exists
    if 'clean_message' not in df_filtered_abuse.columns:
        df_filtered_abuse['clean_message'] = df_filtered_abuse['message'].apply(lambda x: clean_message(str(x)))

    abuse_counts = {}
    for word in hindi_abusive_words:
        count = df_filtered_abuse['clean_message'].str.lower().str.count(word).sum()
        if count > 0: 
            abuse_counts[word] = count

    return abuse_counts

def basic_stats(df_orig, username=None, analyzer_instance=None): # analyzer_instance is no longer needed
    """
    Calculate basic statistics about messages, including sentiment, time analysis,
    most common n-grams (unigrams, bigrams, trigrams), most active period, and visualizations.
    """
    # Prepare user-specific data and initial stats
    df_filtered, positive_msgs, negative_msgs, morning_msgs, \
    midday_msgs, evening_msgs, night_msgs, most_active_period = \
        _prepare_user_data(df_orig, username)

    # Calculate N-grams
    common_unigrams, common_bigrams, common_trigrams = _calculate_ngrams(df_filtered)
    
    # Unique words count (using df_filtered which has clean_message_lower)
    vectorizer = CountVectorizer(stop_words=list(stop_words))
    unique_words_count = 0
    corpus_for_unique_words = df_filtered['clean_message_lower'].dropna()
    if not corpus_for_unique_words.empty:
        try:
            word_matrix = vectorizer.fit_transform(corpus_for_unique_words)
            unique_words_count = len(vectorizer.get_feature_names_out())
        except ValueError: 
            unique_words_count = 0

    # Top 5 Emojis (using df_filtered)
    df_filtered['emojis_list'] = df_filtered['message'].apply(extract_emojis) # Renamed to avoid conflict with df_filtered['emoji'] from basic_cleanup
    all_emojis_list = [emoji_item for sublist in df_filtered['emojis_list'] for emoji_item in sublist]
    top_5_emojis = Counter(all_emojis_list).most_common(5)

    # Average Sentence Length
    df_filtered['sentence_length_basic'] = df_filtered['clean_message'].apply(lambda x: len(nltk.sent_tokenize(str(x))))
    avg_sentence_length = df_filtered['sentence_length_basic'].apply(lambda x: len(str(x).split()) / x if x > 0 else 0).mean()


    # Analyze message timing and get response times
    # Use df_orig here if analyze_message_timing expects the full dataframe before filtering for user
    response_times = analyze_message_timing(df_orig, username) 
    average_response_time = response_times.mean() if not response_times.empty else 0

    # Visualizations (ensure df_filtered is passed, not df_orig for user-specific plots)
    activity_heatmap_base64 = plot_activity_heatmap(df_filtered, username)
    sentiment_distribution_base64 = plot_sentiment_distribution(df_filtered, username)
    wordcloud_base64 = generate_wordcloud(df_filtered, username) # Uses clean_message
    language_complexity_base64 = analyze_language_complexity(df_filtered, username) # Uses clean_message
    response_time_distribution_base64 = plot_response_time_distribution(response_times, username)
    sentiment_over_time_base64 = analyze_sentiment_over_time(df_filtered, username) # Use df_filtered
    emoji_usage_base64 = plot_emoji_usage(df_filtered, username) # Uses 'emojis_list' now
    sentiment_bubble_base64 = plot_sentiment_bubble(df_filtered, username)
    vocabulary_diversity_base64 = plot_vocabulary_diversity(df_filtered, username) # Uses clean_message_lower
    language_complexity_pos_base64 = plot_language_complexity_pos(df_filtered, username)
    user_relationship_graph_base64 = plot_user_relationship_graph(df_orig) # Graph is for all users
    skills_radar_chart_base64 = plot_skills_radar_chart(df_filtered, username) # Uses clean_message
    emotion_over_time_base64 = analyze_emotion_over_time(df_filtered, username)
    most_active_hours_base64 = plot_most_active_hours(df_filtered, username)

    # Analyze behavioral traits (using the standalone function)
    # Pass df_filtered which already has 'clean_message' and 'clean_message_lower'
    behavioral_traits = analyze_behavioral_traits(df_filtered, username)
    behavioral_insights_text = generate_behavioral_insights_text(behavioral_traits, most_active_period, average_response_time)

    # Analyze for Hindi गाली (using the standalone function)
    # Pass df_filtered which already has 'clean_message'
    abuse_counts = analyze_hindi_abuse(df_filtered, username)
    
    abuse_counts_html = "<ul>"
    for word, count in abuse_counts.items():
        abuse_counts_html += f"<li>{word}: {count}</li>"
    abuse_counts_html += "</ul>"

    # Format n-grams as HTML list items
    common_unigrams_html = "".join([f"<li>{word[0]}: {word[1]}</li>" for word in common_unigrams])
    common_bigrams_html = "".join([f"<li>{word[0]}: {word[1]}</li>" for word in common_bigrams])
    common_trigrams_html = "".join([f"<li>{word[0]}: {word[1]}</li>" for word in common_trigrams])

    stats = {
        'Total Messages': len(df_filtered),
        'Total Words': df_filtered['message'].apply(lambda x: len(str(x).split())).sum(),
        'Unique Users': df_filtered['name'].nunique(), # Should be 1 if username is specified
        'Total Emojis': df_filtered['emojicount'].sum() if 'emojicount' in df_filtered.columns else 0,
        'Total URLs': df_filtered['urlcount'].sum() if 'urlcount' in df_filtered.columns else 0,
        'Total YouTube URLs': df_filtered['yturlcount'].sum() if 'yturlcount' in df_filtered.columns else 0,
        'Total Media': df_filtered['mediacount'].sum() if 'mediacount' in df_filtered.columns else 0,
        'Total Edits': df_filtered['editcount'].sum() if 'editcount' in df_filtered.columns else 0,
        'Total Deletes': df_filtered['deletecount'].sum() if 'deletecount' in df_filtered.columns else 0,
        'Average Message Length': df_filtered['mlen'].mean() if 'mlen' in df_filtered.columns else 0,
        'Positive Messages': positive_msgs,
        'Negative Messages': negative_msgs,
        'Morning Messages': morning_msgs,
        'Mid-day Messages': midday_msgs,
        'Evening Messages': evening_msgs,
        'Night Messages': night_msgs,
        'Most Active Period': most_active_period,
        'Unique Words Count': unique_words_count,
        'Common Unigrams': common_unigrams_html, # Store HTML string
        'Common Bigrams': common_bigrams_html,   # Store HTML string
        'Common Trigrams': common_trigrams_html, # Store HTML string
        'Top 5 Emojis': top_5_emojis, # Keep as list of tuples for now, will be processed in analyzer.py
        'Average Sentence Length': avg_sentence_length,
        'Average Response Time': average_response_time,
        'Activity Heatmap': activity_heatmap_base64,
        'Sentiment Distribution': sentiment_distribution_base64,
        'Word Cloud': wordcloud_base64,
        'Language Complexity': language_complexity_base64,
        'Response Time Distribution': response_time_distribution_base64,
        'Sentiment Over Time': sentiment_over_time_base64,
        'Emoji Usage': emoji_usage_base64,
        'Sentiment Bubble': sentiment_bubble_base64,
        'Vocabulary Diversity': vocabulary_diversity_base64,
        'Language Complexity POS': language_complexity_pos_base64,
        'User Relationship Graph': user_relationship_graph_base64,
        'Skills Radar Chart': skills_radar_chart_base64,
        'Behavioral Traits': behavioral_traits, # Dictionary from analyze_behavioral_traits
        'Emotion Over Time': emotion_over_time_base64,
        'Behavioral Insights Text': behavioral_insights_text, # Text from generate_behavioral_insights_text
        'Hindi Abuse Counts': abuse_counts_html, # HTML string from analyze_hindi_abuse counts
        'Most Active Hours': most_active_hours_base64,
    }

    return stats



# tests/test_parser.py

import unittest
import pandas as pd
from datetime import datetime
import os
import tempfile
from whatsapp_analyzer.parser import Parser

class TestParser(unittest.TestCase):

    def setUp(self):
        # Create a temporary file that can be used by multiple tests needing file operations
        self.temp_file_handle, self.temp_file_path = tempfile.mkstemp(suffix=".txt", text=True)
        # Close the handle immediately so it can be opened by other processes or modes
        os.close(self.temp_file_handle)

    def tearDown(self):
        # Clean up the temporary file
        if os.path.exists(self.temp_file_path):
            os.remove(self.temp_file_path)

    # --- Tests for preprocess_lines ---
    def test_preprocess_single_lines(self):
        parser_instance = Parser("dummy.txt") # File path not used for this direct method call
        lines = [
            "10/10/22, 10:00 AM - User1: Message 1\n",
            "10/10/22, 10:01 AM - User2: Message 2\n",
            "10/10/22, 10:02 AM - User1: Message 3\n"
        ]
        processed = parser_instance.preprocess_lines(lines)
        self.assertEqual(len(processed), 3)
        self.assertEqual(processed[0], "10/10/22, 10:00 AM - User1: Message 1")
        self.assertEqual(processed[1], "10/10/22, 10:01 AM - User2: Message 2")
        self.assertEqual(processed[2], "10/10/22, 10:02 AM - User1: Message 3")

    def test_preprocess_multiline_messages(self):
        parser_instance = Parser("dummy.txt")
        lines = [
            "10/10/22, 10:00 AM - User1: This is the first line\n",
            "of a multiline message.\n",
            "10/10/22, 10:01 AM - User2: This is a single line message.\n",
            "10/10/22, 10:02 AM - User1: Another multiline\n",
            "message here\n",
            "and here.\n"
        ]
        processed = parser_instance.preprocess_lines(lines)
        self.assertEqual(len(processed), 3)
        self.assertEqual(processed[0], "10/10/22, 10:00 AM - User1: This is the first line of a multiline message.")
        self.assertEqual(processed[1], "10/10/22, 10:01 AM - User2: This is a single line message.")
        self.assertEqual(processed[2], "10/10/22, 10:02 AM - User1: Another multiline message here and here.")

    def test_preprocess_empty_input(self):
        parser_instance = Parser("dummy.txt")
        lines = []
        processed = parser_instance.preprocess_lines(lines)
        self.assertEqual(len(processed), 0)

    # --- Tests for parse_chat_data ---
    def test_parse_chat_data_simple(self):
        chat_content = [
            "10/10/22, 10:00 AM - User1: Hello there!\n",
            "10/10/22, 10:01 AM - User2: Hi User1.\n",
            "10/10/22, 10:02 AM - User2 changed this group's icon\n" 
        ]
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            f.writelines(chat_content)

        parser_instance = Parser(self.temp_file_path)
        df = parser_instance.parse_chat_data()

        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(len(df), 3)
        
        self.assertEqual(df.iloc[0]['t'], datetime(2022, 10, 10, 10, 0, 0))
        self.assertEqual(df.iloc[0]['name'], "User1")
        self.assertEqual(df.iloc[0]['message'], "Hello there!")

        self.assertEqual(df.iloc[1]['t'], datetime(2022, 10, 10, 10, 1, 0))
        self.assertEqual(df.iloc[1]['name'], "User2")
        self.assertEqual(df.iloc[1]['message'], "Hi User1.")
        
        self.assertEqual(df.iloc[2]['t'], datetime(2022, 10, 10, 10, 2, 0))
        self.assertEqual(df.iloc[2]['name'], "System") 
        self.assertEqual(df.iloc[2]['message'], "User2 changed this group's icon")


    def test_parse_chat_data_multiline(self):
        chat_content = [
            "10/10/22, 11:00 AM - UserA: First part of message.\n",
            "Second part of message.\n",
            "Third part.\n",
            "10/10/22, 11:05 AM - UserB: Another message.\n"
        ]
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            f.writelines(chat_content)
        
        parser_instance = Parser(self.temp_file_path)
        df = parser_instance.parse_chat_data()

        self.assertEqual(len(df), 2)
        self.assertEqual(df.iloc[0]['message'], "First part of message. Second part of message. Third part.")
        self.assertEqual(df.iloc[0]['name'], "UserA")
        self.assertEqual(df.iloc[1]['name'], "UserB")

    def test_parse_chat_data_system_messages(self):
        chat_content = [
            "01/01/23, 12:01 AM - User1 joined using this group's invite link\n",
            "01/01/23, 12:02 AM - You created group \"Test Group\"\n",
            "01/01/23, 12:03 AM - User2: Hello all!\n"
        ]
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            f.writelines(chat_content)

        parser_instance = Parser(self.temp_file_path)
        df = parser_instance.parse_chat_data()
        
        self.assertEqual(len(df), 3)
        self.assertEqual(df.iloc[0]['name'], "System")
        self.assertEqual(df.iloc[0]['message'], "User1 joined using this group's invite link")
        self.assertEqual(df.iloc[1]['name'], "System")
        self.assertEqual(df.iloc[1]['message'], "You created group \"Test Group\"")
        self.assertEqual(df.iloc[2]['name'], "User2")
        self.assertEqual(df.iloc[2]['message'], "Hello all!")

    def test_parse_chat_data_invalid_lines(self):
        chat_content = [
            "This is a completely invalid line.\n", 
            "10/10/22, 10:00 AM - User1: Valid message 1.\n",
            "Another invalid line, no timestamp.\n", 
            "10/10/22, 10:01 AM - User2: Valid message 2.\n",
            "This line belongs to message 2.\n", 
            "Yet another invalid line.\n" 
        ]
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            f.writelines(chat_content)

        parser_instance = Parser(self.temp_file_path)
        df = parser_instance.parse_chat_data()
        
        self.assertEqual(len(df), 2) 
        self.assertEqual(df.iloc[0]['message'], "Valid message 1. Another invalid line, no timestamp.")
        self.assertEqual(df.iloc[1]['message'], "Valid message 2. This line belongs to message 2. Yet another invalid line.")


    def test_parse_chat_data_file_not_found(self):
        parser_instance = Parser("non_existent_file.txt")
        with self.assertRaises(FileNotFoundError):
            parser_instance.parse_chat_data()

    def test_parse_chat_data_empty_file(self):
        with open(self.temp_file_path, 'w', encoding='utf-8') as f:
            pass 
        
        parser_instance = Parser(self.temp_file_path)
        df = parser_instance.parse_chat_data()
        
        self.assertIsInstance(df, pd.DataFrame)
        self.assertTrue(df.empty)

if __name__ == "__main__":
    unittest.main()


# whatsapp_analyzer/utils.py
import regex
import emoji
import pandas as pd
import numpy as np
import re
import itertools
from .parser import Parser
from .constants import skill_keywords, hindi_abusive_words, stop_words

URL_PATTERN = r"(https?://\S+)"
YOUTUBE_PATTERN = r"(https?://youtu(\.be|be\.com)\S+)"

ANIMAL_NAMES = ["Panda", "Shark", "Zebra", "Lion", "Tiger", "Bear", "Eagle", "Wolf", "Fox", "Deer"]

def get_emojis(text):
    emoji_list = []
    data = regex.findall(r"\X", text)
    for word in data:
        if any(char in emoji.EMOJI_DATA for char in word):
            emoji_list.append(word)
    return emoji_list

def get_urls(text):
    url_list = regex.findall(URL_PATTERN, text)
    return url_list

def get_yturls(text):
    url_list = regex.findall(YOUTUBE_PATTERN, text)
    return url_list

def _extract_datetime_features(df):
    """Extracts date and time based features from the 'date_time' column."""
    df["date"] = df["date_time"].dt.date
    df["year"] = df["date_time"].dt.year
    df["month"] = df["date_time"].dt.month.astype(str).str.zfill(2)
    df["day"] = df["date_time"].dt.day

    df["dayn"] = df["date_time"].dt.day_name().astype("category")
    df["monthn"] = df["date_time"].dt.month_name()

    df["doy"] = df["date_time"].dt.dayofyear
    df["dow"] = df["date_time"].dt.dayofweek
    df["woy"] = df["date_time"].dt.isocalendar().week.astype(str) # Ensure woy is string for yw concatenation
    df["time"] = df["date_time"].dt.time
    df["hour"] = df["date_time"].dt.hour
    df["min"] = df["date_time"].dt.minute
    df["hm"] = df["hour"] + round(df["min"] / 60, 2)

    df["ym"] = df["year"].astype(str) + "-" + df["month"]
    df["yw"] = df["year"].astype(str) + "-" + df["woy"]
    df["yd"] = df["year"].astype(str) + "-" + df["doy"].astype(str).str.zfill(3) # Ensure doy is string and padded
    df["md"] = df["monthn"].astype(str) + "-" + df["day"].astype(str).str.zfill(2) # Ensure day is string and padded
    return df

def _extract_message_content_features(df):
    """Extracts features from the message content."""
    df["mlen"] = df["message"].str.len()
    df["emoji"] = df["message"].apply(get_emojis)
    df["emojicount"] = df["emoji"].str.len()
    df["urls"] = df["message"].apply(get_urls)
    df["urlcount"] = df["urls"].str.len()
    df["yturls"] = df["message"].apply(get_yturls)
    df["yturlcount"] = df["yturls"].str.len()
    return df

def _extract_message_type_counts(df):
    """Extracts counts of media, edited, and deleted messages."""
    df["mediacount"] = np.where(df["message"] == "<Media omitted>", 1, 0)
    df["editcount"] = np.where(
        df["message"].str.contains("<This message was edited>"), 1, 0
    )
    df["deletecount"] = np.where(
        (
            (df["message"] == "This message was deleted")
            | (df["message"] == "You deleted this message")
        ),
        1,
        0,
    )
    return df

def df_basic_cleanup(df):
    """
    Performs basic data cleaning and feature engineering on the DataFrame.
    Orchestrates calls to helper functions to extract features.
    """
    # Ensure 't' column exists and is the primary source for 'date_time'
    if 't' not in df.columns:
        raise ValueError("Input DataFrame must contain a 't' column for timestamp information.")
    
    df["date_time"] = pd.to_datetime(df["t"], errors='coerce')
    
    # Drop rows where 'date_time' could not be parsed, if any
    df.dropna(subset=['date_time'], inplace=True)

    df = _extract_datetime_features(df)
    df = _extract_message_content_features(df)
    df = _extract_message_type_counts(df)

    df.drop("t", inplace=True, axis=1)
    
    # Define the final order of columns
    final_columns_order = [
            "date_time",
            "date",
            "year",
            "month",
            "monthn",
            "day",
            "dayn",
            "woy",
            "doy",
            "dow",
            "ym",
            "yw",
            "yd",
            "md",
            "time",
            "hour",
            "min",
            "hm",
            "name",
            "message",
            "mlen",
            "emoji",
            "emojicount",
            "urls",
            "urlcount",
            "yturls",
            "yturlcount",
            "mediacount",
            "editcount",
            "deletecount",
        ]
    # The extra ']' was here, it has been removed.
    # Filter out any columns not in the original DataFrame to handle cases where some columns might be missing
    # This also ensures that if new columns are added by helpers but not in final_columns_order, they are dropped.
    existing_columns = [col for col in final_columns_order if col in df.columns]
    df = df[existing_columns]
    
    return df

def anonymize(input_chat_path: str, output_chat_path: str):
    """Anonymizes usernames in a WhatsApp chat file and saves it to a new file."""
    username_to_anonymous_map = {}
    user_id_counter = 0
    animal_cycle = itertools.cycle(ANIMAL_NAMES)
    parsed_lines_data = []

    # Define regex patterns for parsing chat lines
    # Pattern for user messages: captures date, time, author, and message
    # e.g., "20/03/2023, 10:00 AM - Alice: Hello Bob"
    user_message_pattern = re.compile(
        r"(\d{1,2}/\d{1,2}/\d{2,4}),\s*(\d{1,2}:\d{2}(?:\s*(?:AM|PM|am|pm))?)\s*-\s*([^:]+?):\s*(.*)"
    )
    # Pattern for system messages or events that don't have a specific author mentioned with a colon
    # e.g., "20/03/2023, 10:05 AM - You deleted this message" or "20/03/2023, 10:00 AM - System: Alice added Bob" (where "System: " is part of the message)
    # For anonymization purposes, we mostly care about the structure to extract date/time and differentiate from user messages.
    # The author for these will effectively be "System" or the line is kept as is if no author is truly identifiable.
    system_event_pattern = re.compile(
        r"(\d{1,2}/\d{1,2}/\d{2,4}),\s*(\d{1,2}:\d{2}(?:\s*(?:AM|PM|am|pm))?)\s*-\s*(.*)"
    )

    username_to_anonymous_map = {}
    user_id_counter = 0
    animal_cycle = itertools.cycle(ANIMAL_NAMES)
    parsed_lines_data = []

    try:
        with open(input_chat_path, "r", encoding="utf-8") as f:
            for line in f:
                stripped_line = line.strip()
                if not stripped_line:  # Skip empty lines
                    parsed_lines_data.append(stripped_line)
                    continue

                parsed_info = None
                user_match = user_message_pattern.match(stripped_line)
                if user_match:
                    date_str, time_str, author_str, message_str = user_match.groups()
                    parsed_info = {
                        "date": date_str.strip(),
                        "time": time_str.strip(),
                        "author": author_str.strip(),
                        "message": message_str.strip()
                    }
                else:
                    system_match = system_event_pattern.match(stripped_line)
                    if system_match:
                        date_str, time_str, message_str = system_match.groups()
                        # If the message content starts with "System: ", treat "System" as author
                        # This is a heuristic; actual system messages might vary.
                        # The main goal here is to differentiate from user messages for anonymization.
                        # The logic below (`author.lower() != "system"`) handles "System" authors correctly.
                        # For lines like "You deleted this message", the author is effectively System.
                        # Let's assume author is "System" for all system_event_pattern matches for now.
                        # The key is that these names won't be added to username_to_anonymous_map unless they are actual users.
                        parsed_info = {
                            "date": date_str.strip(),
                            "time": time_str.strip(),
                            "author": "System", # Default for system events
                            "message": message_str.strip()
                        }
                        # A more robust way to check if it's a system message for anonymization purposes:
                        # if the message_str itself contains a colon, it might be a user message that was
                        # misparsed by user_message_pattern (e.g. unusual characters in username).
                        # However, the provided regex for user_message_pattern is greedy for author.
                        # For now, this simplified approach for system messages.

                if parsed_info and parsed_info.get('author'):
                    author = parsed_info['author']
                    # Check if author is 'System' (case-insensitive) or already mapped
                    if author.lower() != "system" and author not in username_to_anonymous_map:
                        user_id_counter += 1
                        anonymous_name = f"user_{user_id_counter}_{next(animal_cycle)}"
                        username_to_anonymous_map[author] = anonymous_name
                    parsed_lines_data.append(parsed_info)
                else:
                    # If line is not parsed or has no author, store the original line
                    parsed_lines_data.append(stripped_line)
    
    except FileNotFoundError:
        print(f"Error: Input file '{input_chat_path}' not found.")
        return
    except Exception as e:
        print(f"An error occurred: {e}") # General exception handling
        return

    # For debugging purposes (to be replaced with file writing logic later)
    # Remove previous debugging print statements
    # print("Username to Anonymous Map:", username_to_anonymous_map)
    # print("\nParsed Lines Data:")
    # for item in parsed_lines_data:
    #     print(item)

    try:
        with open(output_chat_path, "w", encoding="utf-8") as f:
            for item in parsed_lines_data:
                if isinstance(item, dict):
                    author = item['author']
                    # Ensure message is not None, default to empty string if it is
                    message = item.get('message', '') 

                    if author.lower() == "system":
                        display_author = author 
                    else:
                        display_author = username_to_anonymous_map.get(author, author)
                    
                    reconstructed_line = f"{item['date']}, {item['time']} - {display_author}: {message}"
                    f.write(reconstructed_line + '\n')
                else:
                    # item is an original string (unparsed line or empty line)
                    f.write(item + '\n')
        print(f"Anonymized chat saved to {output_chat_path}")
    except IOError as e:
        print(f"Error writing to output file '{output_chat_path}': {e}")
    except Exception as e:
        print(f"An unexpected error occurred during file writing: {e}")


import pandas as pd
from nltk.corpus import stopwords

# Now add the variables at the end of the file
custom_hinglish_stopwords = set([
    '<media omitted>', 'media', 'omitted', 'bhai', 'hai', 'kya', 'ka', 'ki', 'ke', 'h', 'nahi', 'haan', 'ha',
    'to', 'ye', 'ho', 'na', 'ko', 'se', 'me', 'mai', 'mera', 'apna', 'tum', 'mujhe', 'jo',
    'bhi', 'nhi', 'hi', 'rha', 'tha', 'hain', 'abhi', 'kr', 'rha', 'thi', 'kar', 'karna',
    'raha', 'rahe', 'gaya', 'gayi', 'kyun', 'acha', 'lo', 'pe', 'kaun', 'tumhare', 'unki',
    'message', 'wo', 'koi', 'aa', 'le', 'ek', 'mei', 'lab', 'aur', 'kal', 'sab', 'us', 'un',
    'hum', 'kab', 'ab', 'par', 'kaise', 'unka', 'ap', 'mere', 'tere', 'kar', 'deleted', 'hun', 'hu', 'ne',
    'tu', 'ya', 'edited'
])

# Combine NLTK stopwords with custom Hinglish stopwords
stop_words = set(stopwords.words('english')).union(custom_hinglish_stopwords)

skill_keywords = {
    'communication': [
        'talk', 'discuss', 'share', 'convey', 'express', 'message', 'articulate',
        'explain', 'correspond', 'batana', 'samjhana', 'bataana', 'baat', 'dono',
        'tell', 'suno', 'dikhana', 'bol', 'bolna', 'likhna', 'likh', 'samaj',
        'sun', 'keh', 'kehna', 'padhana', 'janana', 'jan', 'vyakth karna', 'samjhao',
        'dekh', 'dekhna','sunana','samvad','guftgu','prastut','izhaar','pragatikaran','viniyog'
    ],
    'leadership': [
        'guide', 'manage', 'lead', 'organize', 'direct', 'influence', 'motivate',
        'inspire', 'leadership', 'rahnumai', 'neta banna', 'lead karna', 'manage karna',
        'prabhaavit karna', 'dhikhaana', 'aguvai', 'nirdeshan', 'niyantran',
        'prabandhak', 'netritvakarta', 'pravartak', 'diksha', 'dekhrekh','chalana','niyantran karna'
    ],
    'problem_solving': [
        'solve', 'resolve', 'analyze', 'figure', 'fix', 'improve', 'optimize',
        'address', 'determine', 'solve karna', 'masla suljhna', 'improve karna',
        'sahi karna', 'thik karna', 'dhoondhna', 'hal karna', 'samadhan', 'niptara',
        'sudharna', 'behtar', 'anukulan', 'nirdharan',  'gyat','thik karna',
        'samadhan sochna', 'samadhan ka upyog', 'samadhanikaran', 'samadhan dena'
    ],
    'technical': [
        'code', 'program', 'algorithm', 'software', 'hardware', 'system', 'network',
        'database', 'debug', 'coding', 'programming', 'debugging', 'networking',
        'computer', 'server', 'database kaam', 'tech', 'cloud', 'app', 'automation',
        'hardware ki setting', 'takniki', 'praudyogiki', 'yantrik', 'abhikalpan',
        'karya', 'karya pranali', 'vidhi', 'tantra','upkaran', 'samagri', 'sangathan', 
        'sanchar', 'aankda', 'soochi', 'doshal', 'tantrik', 'vigyan', 'software vikas',
        'hardware vikas', 'network sthapana', 'database prabandhan', 'debug karna',
        "lower bound", "upper bound", "time complexity", "space complexity", "algorithm design",
        "estimation", "nvidia", "detection", "classification", "regression", "prediction",
    ],
    'teamwork': [
        'collaborate', 'cooperate', 'coordinate', 'assist', 'support', 'together',
        'contribute', 'participate', 'teamwork', 'saath kaam karna', 'mil jul kar kaam',
        'sath dena', 'madad karna', 'sahyog karna', 'support karna', 'cooperate karna',
        'milkar', 'sath', 'sahkarya', 'sajha', 'sahkari', 'sahbhaagi', 'samudaayik', 'ekjut',
        'sammilit', 'gatbandhan','sahyog dena', "bhardo", "kardo", "bhejdo"
    ]
}

hindi_abusive_words = [
        'chutiya', 'gandu', 'bhosdike', 'bhadwe', 'madarchod', 'behenchod', 'randi',
        'laude', 'chut', 'harami', 'kutta', 'kutiya', 'suar', 'hijra', 'gaand', 'tatte',
        'jhat', 'bhosdi', 'bhadwa', 'chinal', 'chakka', 'behen ke laude', 'maa ke laude',
        'baap ke laude', 'bhosdiwala', 'bhosdiwali', 'gandu ke aulad', 'gandi aulad',
        'harami aulad', 'gandu sala', 'chutiya sala', 'bhosdike sala', 'madarchod sala',
        'gandi maa ka', 'gandi maa ki', 'gandu maa ka', 'gandu maa ki', 'chutiya maa ka',
        'chutiya maa ki', 'madarchod maa ka', 'madarchod maa ki', 'madarchod bhai',
        'madarchod bahen', 'bhosdike bhai', 'bhosdike bahen', 'chutiya bhai', 'chutiya bahen',
        'gandu bhai', 'gandu bahen', 'harami bhai', 'harami bahen', 'bhadwe bhai', 'bhadwe bahen',
        'bsdiwala', 'iski maka', 'betichod', "gand", "bc", "mc", "madar", "bkl", "bkl", "bkl", "bkl",]

html_template = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>WhatsApp Chat Analysis - {name}</title>
    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css">
    <link rel="stylesheet" href="https://fonts.googleapis.com/css2?family=Roboto:wght@400;700&display=swap">
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.1.1/css/all.min.css">
    <style>
    
body {{
            font-family: 'Roboto', Arial, sans-serif;
            background-color: #f4f4f4;
            color: #333;
            line-height: 1.6;
        }}
        .container {{
            max-width: 1100px;
            margin: auto;
            padding: 20px;
        }}
        .header {{
            background-color: #075e54;
            color: #fff;
            padding: 20px;
            text-align: center;
            border-radius: 10px 10px 0 0;
        }}
        .profile-card {{
            background-color: #fff;
            border-radius: 10px;
            box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
            text-align: center;
            padding: 20px;
        }}
        .profile-img {{
            border-radius: 50%;
            width: 120px;
            height: 120px;
            margin: 0 auto 15px;
            object-fit: cover;
        }}
        .username {{
            font-size: 1.8rem;
            font-weight: 700;
            margin-bottom: 10px;
        }}
        .status {{
            font-size: 1.1rem;
            color: #4CAF50;
            margin-bottom: 15px;
        }}
        .location, .social-links {{
            font-size: 1rem;
            color: #555;
            margin-bottom: 15px;
        }}
        .social-links a {{
            margin: 0 10px;
            color: #fff;
            padding: 8px 15px;
            border-radius: 5px;
            text-decoration: none;
        }}
        .social-links a.facebook {{ background-color: #3b5998; }}
        .social-links a.instagram {{ background-color: #e4405f; }}
        .user-report {{
            background-color: #fff;
            border-radius: 10px;
            box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
            padding: 20px;
            margin-top: 20px;
        }}
        .section-title {{
            color: #075e54;
            font-size: 1.5rem;
            margin-bottom: 15px;
        }}
        .table {{
            background-color: #fff;
            border-radius: 10px;
            overflow: hidden;
            box-shadow: 0 2px 5px rgba(0, 0, 0, 0.05);
        }}
        .table th {{
            background-color: #075e54;
            color: #fff;
        }}
        .table th, .table td {{
            padding: 12px 15px;
            text-align: left;
        }}
        .footer {{
            background-color: #075e54;
            color: #fff;
            text-align: center;
            padding: 20px;
            margin-top: 20px;
            border-radius: 0 0 10px 10px;
        }}
        .footer a {{
            color: #fff;
            text-decoration: none;
        }}
        .footer a:hover {{
            text-decoration: underline;
        }}
        .emoji {{
            font-size: 1.2rem;
            font-family: "Segoe UI Emoji", "Apple Color Emoji", 'Roboto', Arial, sans-serif;
        }}
        .visualization {{
            margin-top: 20px;
            text-align: center;
        }}
        .visualization img {{
            max-width: 100%;
            height: auto;
            border-radius: 10px;
            box-shadow: 0 4px 8px rgba(0, 0, 0, 0.1);
        }}
        .insights {{
            margin-top: 20px;
            padding: 15px;
            background-color: #f0f0f0;
            border-radius: 10px;
        }}
        .insights h4 {{
            color: #075e54;
            margin-bottom: 10px;
        }}
        .insights p {{
            font-size: 0.9rem;
            line-height: 1.5;
        }}
    </style>
</head>
<body>
    <div class="container">
        <header class="header">
            <h1>WhatsApp Chat Analysis - {name}</h1>
        </header>
        <div class="row mt-4">
            <div class="col-md-3">
                <div class="profile-card">
                    <img src="https://via.placeholder.com/150" alt="{name}'s Profile Picture" class="profile-img">
                    <h3 class="username">{name}</h3>
                    <p class="status">Active User</p>
                    <p class="location"><i class="fas fa-map-marker-alt"></i> Location: New Delhi</p>
                    
                </div>
            </div>
            <div class="col-md-9">
                <div class="user-report">
                    <div class="section">
                        <h2 class="section-title">User Stats</h2>
                        <table class="table table-striped">
                            <tr><th>Total Messages</th><td>{Total Messages}</td></tr>
                            <tr><th>Total Words</th><td>{Total Words}</td></tr>
                            <tr><th>Unique Users</th><td>{Unique Users}</td></tr>
                            <tr><th>Total Emojis</th><td>{Total Emojis}</td></tr>
                            <tr><th>Top 5 Emojis</th><td class="emoji">{top_5_emojis_html}</td></tr>
                            <tr><th>Total URLs</th><td>{Total URLs}</td></tr>
                            <tr><th>Total YouTube URLs</th><td>{Total YouTube URLs}</td></tr>
                            <tr><th>Total Media</th><td>{Total Media}</td></tr>
                            <tr><th>Total Edits</th><td>{Total Edits}</td></tr>
                            <tr><th>Total Deletes</th><td>{Total Deletes}</td></tr>
                            <tr><th>Average Message Length</th><td>{Average Message Length:.2f}</td></tr>
                            <tr><th>Average Sentence Length</th><td>{Average Sentence Length:.2f}</td></tr>
                            <tr><th>Positive Messages</th><td>{Positive Messages}</td></tr>
                            <tr><th>Negative Messages</th><td>{Negative Messages}</td></tr>
                            <tr><th>Morning Messages</th><td>{Morning Messages}</td></tr>
                            <tr><th>Mid-day Messages</th><td>{Mid-day Messages}</td></tr>
                            <tr><th>Evening Messages</th><td>{Evening Messages}</td></tr>
                            <tr><th>Night Messages</th><td>{Night Messages}</td></tr>
                            <tr><th>Most Active Period</th><td>{Most Active Period}</td></tr>
                            <tr><th>Unique Words Count</th><td>{Unique Words Count}</td></tr>
                            <tr><th>Average Response Time (minutes)</th><td>{Average Response Time:.2f}</td></tr>
                        </table>
                    </div>
                    <div class="section">
                        <h2 class="section-title">Common Words</h2>
                        <h3>Unigrams</h3>
                        <ul>
                            {Common Unigrams}
                        </ul>
                        <h3>Bigrams</h3>
                        <ul>
                            {Common Bigrams}
                        </ul>
                        <h3>Trigrams</h3>
                        <ul>
                            {Common Trigrams}
                        </ul>
                        <h3>Hindi abuse</h3>
                        <ul>
                            {Hindi Abuse Counts}
                        </ul>
                    </div>
                    <div class="section">
                        <h2 class="section-title">Visualizations</h2>
                              
                        <div class="visualization">
                            <h4>Most Active Hours</h4>
                            <img src="data:image/png;base64,{Most Active Hours}" alt="Most Active Hours">
                        </div>

    
                        <div class="visualization">
                            <h4>Activity Heatmap</h4>
                            <img src="data:image/png;base64,{Activity Heatmap}" alt="Activity Heatmap">
                        </div>
                        <div class="visualization">
                            <h4>Response Time Distribution</h4>
                            <img src="data:image/png;base64,{Response Time Distribution}" alt="Response Time Distribution">
                        </div>
                        <div class="visualization">
                            <h4>Sentiment Over Time</h4>
                            <img src="data:image/png;base64,{Sentiment Over Time}" alt="Sentiment Over Time">
                        </div>
                        <div class="visualization">
                            <h4>Emoji Usage</h4>
                            <img src="data:image/png;base64,{Emoji Usage}" alt="Emoji Usage">
                        </div>
                        <div class="visualization">
                            <h4>Sentiment Distribution</h4>
                            <img src="data:image/png;base64,{Sentiment Distribution}" alt="Sentiment Distribution">
                        </div>
                        <div class="visualization">
                            <h4>Sentiment (Bubble)</h4>
                            <img src="data:image/png;base64,{Sentiment Bubble}" alt="Sentiment Bubble">
                        </div>
                        <div class="visualization">
                            <h4>Vocabulary Diversity</h4>
                            <img src="data:image/png;base64,{Vocabulary Diversity}" alt="Vocabulary Diversity">
                        </div>
                        <div class="visualization">
                            <h4>Language Complexity</h4>
                            <img src="data:image/png;base64,{Language Complexity}" alt="Language Complexity">
                        </div>
                        <div class="visualization">
                            <h4>Language Complexity (POS)</h4>
                            <img src="data:image/png;base64,{Language Complexity POS}" alt="Language Complexity POS">
                        </div>
                        <div class="visualization">
                            <h4>User Relationship Graph</h4>
                            <img src="data:image/png;base64,{User Relationship Graph}" alt="User Relationship Graph">
                        </div>
                        <div class="visualization">
                            <h4>Skills Radar Chart</h4>
                            <img src="data:image/png;base64,{Skills Radar Chart}" alt="Skills Radar Chart">
                        </div>
                        <div class="visualization">
                            <h4>Emotion Trends (Time Series)</h4>
                            <img src="data:image/png;base64,{Emotion Over Time}" alt="Emotion Over Time">
                        </div>
                        <div class="visualization">
                            <h4>Word Cloud</h4>
                            <img src="data:image/png;base64,{Word Cloud}" alt="Word Cloud">
                        </div>
                    </div>
                    <div class="section">
                        <h2 class="section-title">Behavioral Insights</h2>
                        <div class="insights">
                            {Behavioral Insights Text}
                        </div>
                    </div>
                </div>
            </div>
        </div>
        <footer class="footer">
            <p>Generated with <i class="fas fa-heart"></i> by WhatsApp Analyzer</p>
            <p><a href="https://github.com/gauravmeena0708/k" target="_blank"><i class="fab fa-github"></i> Visit the Project</a></p>
        </footer>
    </div>
</body>
</html>
"""



import os # For os.path.join and os.makedirs if out_dir needs creation by analyzer
import matplotlib.pyplot as plt # For plt.rcParams
import matplotlib.font_manager as fm # For font management
import warnings # For warnings.warn

from .analyzer import WhatsAppAnalyzer # Import the main class

# Get all installed font names
available_fonts = {fm.FontProperties(fname=fp).get_name() for fp in fm.findSystemFonts(fontext='ttf')} # Added fontext for robustness

# Add an emoji-compatible font if available
emoji_fonts = ["Segoe UI Emoji", "Apple Color Emoji", "Noto Color Emoji"] # Added Noto
selected_font = None

for font in emoji_fonts:
    if font in available_fonts:
        selected_font = font
        break

if selected_font:
    plt.rcParams["font.family"] = [selected_font, "Roboto", "DejaVu Sans", "sans-serif"]
else:
    warnings.warn(
        "No emoji-compatible font found. Install 'Segoe UI Emoji', 'Apple Color Emoji', or 'Noto Color Emoji' for full emoji support."
    )
    plt.rcParams["font.family"] = ["Roboto", "DejaVu Sans", "sans-serif"]

def main():
    """Main function to run the WhatsApp chat analysis."""

    chat_file = "../data/whatsapp_chat.txt"  # Replace with your chat file path
    output_dir = "../data"                  # Specify output directory for reports

    # Initialize WhatsAppAnalyzer
    analyzer = WhatsAppAnalyzer(chat_file=chat_file, out_dir=output_dir)

    # Generate reports for all users found in the chat file
    # The generate_report method handles iterating through users and saving files.
    analyzer.generate_report()

    print(f"All reports have been generated and saved in the '{output_dir}' directory.")

if __name__ == "__main__":
    main()


# setup.py

from setuptools import setup, find_packages

setup(
    name='whatsapp-groupchat-analyzer',
    version='1.0.7',  # Increment the version number
    author="Gaurav Meena",
    author_email="gauravmeena0708@gmail.com",
    description="A Python package for analyzing WhatsApp group chats.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/gauravmeena0708/whatsapp-analyzer",  # Your project URL
    packages=find_packages(),
    install_requires=[
        "seaborn>=0.9",
        "textblob",
        "emoji",
        "networkx",
        "matplotlib",
        "wordcloud",
        "nltk",
        "scikit-learn"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # Choose a license
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",  # Minimum Python version
)



# tests/test_analysis_utils.py

import unittest
import pandas as pd
import numpy as np
from datetime import datetime
import os
import tempfile

from whatsapp_analyzer.analysis_utils import analyze_message_timing, basic_stats
from whatsapp_analyzer.utils import df_basic_cleanup
from whatsapp_analyzer.parser import Parser
# For basic_stats, TextBlob is an indirect dependency for sentiment.
# NLTK resources (punkt, stopwords, vader_lexicon, averaged_perceptron_tagger, wordnet)
# are also expected to be available for basic_stats to run fully.

class TestAnalysisUtils(unittest.TestCase):

    def setUp(self):
        self.temp_file_handle, self.temp_file_path = tempfile.mkstemp(suffix=".txt", text=True)
        os.close(self.temp_file_handle)

    def tearDown(self):
        if os.path.exists(self.temp_file_path):
            os.remove(self.temp_file_path)

    def _create_and_parse_chat_file(self, chat_lines_list_of_strings):
        """Helper to create a chat file, parse it, and run df_basic_cleanup."""
        with open(self.temp_file_path, "w", encoding="utf-8") as f:
            for line in chat_lines_list_of_strings:
                f.write(line + "\n")
        parser = Parser(self.temp_file_path)
        raw_df = parser.parse_chat_data()
        self.assertFalse(raw_df.empty, "Parser returned an empty DataFrame from test data.")
        # 't' column must exist for df_basic_cleanup
        self.assertIn('t', raw_df.columns, "Raw DataFrame must contain 't' column.")
        return df_basic_cleanup(raw_df.copy())

    # --- Tests for analyze_message_timing ---

    def test_analyze_message_timing_single_user(self):
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Message 1",
            "10/10/22, 10:05 AM - User1: Message 2", # 5 min diff
            "10/10/22, 10:00 AM - User2: U2 Message 1",
            "10/10/22, 10:10 AM - User1: Message 3", # 5 min diff from User1's last
            "10/10/22, 10:07 AM - User2: U2 Message 2"  # 7 min diff
        ]
        df = self._create_and_parse_chat_file(chat_lines)
        response_times_user1 = analyze_message_timing(df, username='User1')
        
        self.assertIsInstance(response_times_user1, pd.Series)
        self.assertEqual(len(response_times_user1), 2)
        pd.testing.assert_series_equal(response_times_user1, pd.Series([5.0, 5.0]), check_names=False, check_index=False)

    def test_analyze_message_timing_all_users(self):
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Message 1",
            "10/10/22, 10:05 AM - User1: Message 2", # User1: 5 min
            "10/10/22, 10:00 AM - User2: U2 Message 1",
            "10/10/22, 10:07 AM - User2: U2 Message 2", # User2: 7 min
            "10/10/22, 10:10 AM - User1: Message 3"  # User1: 5 min
        ]
        df = self._create_and_parse_chat_file(chat_lines)
        response_times_all = analyze_message_timing(df) # No username
        
        self.assertIsInstance(response_times_all, pd.Series)
        # Expecting 2 from User1, 1 from User2
        self.assertEqual(len(response_times_all), 3) 
        # Values will be [5.0, 5.0, 7.0] but order depends on groupby-diff-dropna sequence
        expected_values = sorted([5.0, 5.0, 7.0])
        actual_values = sorted(response_times_all.tolist())
        self.assertEqual(actual_values, expected_values)


    def test_analyze_message_timing_one_message_user(self):
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Only one message",
            "10/10/22, 10:05 AM - User2: U2 Message 1",
            "10/10/22, 10:10 AM - User2: U2 Message 2" 
        ]
        df = self._create_and_parse_chat_file(chat_lines)
        response_times_user1 = analyze_message_timing(df, username='User1')
        
        self.assertIsInstance(response_times_user1, pd.Series)
        self.assertTrue(response_times_user1.empty)

    def test_analyze_message_timing_known_gaps(self):
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Msg1",
            "10/10/22, 10:05 AM - User1: Msg2", # 5 min after Msg1
            "10/10/22, 10:20 AM - User1: Msg3"  # 15 min after Msg2
        ]
        df = self._create_and_parse_chat_file(chat_lines)
        response_times_user1 = analyze_message_timing(df, username='User1')
        
        self.assertIsInstance(response_times_user1, pd.Series)
        self.assertEqual(len(response_times_user1), 2)
        pd.testing.assert_series_equal(response_times_user1, pd.Series([5.0, 15.0]), check_names=False, check_index=False)

    # --- Tests for basic_stats (selected checks) ---

    def test_basic_stats_core_metrics(self):
        chat_lines = [
            "10/10/22, 09:00 AM - User1: Good morning! Feeling happy today. 😊",
            "10/10/22, 09:05 AM - User1: Let's make this a great day.",
            "10/10/22, 09:10 AM - User2: Sad news from my side.",
            "10/10/22, 09:15 AM - User1: Oh no, what happened? Bad news is not good."
        ]
        clean_df = self._create_and_parse_chat_file(chat_lines)
        # Ensure 'User1' exists in the parsed data
        self.assertTrue("User1" in clean_df['name'].unique(), "User1 should be in the DataFrame")

        stats_dict_user1 = basic_stats(clean_df.copy(), username='User1')

        self.assertEqual(stats_dict_user1['Total Messages'], 3)
        # "Good morning! Feeling happy today. 😊" -> 5 words (emoji not counted as word by split)
        # "Let's make this a great day." -> 7 words
        # "Oh no, what happened? Bad news is not good." -> 9 words
        self.assertEqual(stats_dict_user1['Total Words'], 5 + 7 + 9) 
        
        # Sentiment: "happy" positive, "good" positive. "sad" negative, "bad" negative.
        # User1 messages:
        # 1: "Good morning! Feeling happy today. 😊" (Positive due to "Good", "happy")
        # 2: "Let's make this a great day." (Positive due to "great")
        # 3: "Oh no, what happened? Bad news is not good." (Negative due to "Bad", "not good")
        # Note: TextBlob's default sentiment might vary. These are simple checks.
        # Requires TextBlob to be installed and vader_lexicon for TextBlob if not already downloaded by NLTK.
        self.assertGreaterEqual(stats_dict_user1['Positive Messages'], 2, "Expected at least 2 positive messages for User1")
        self.assertGreaterEqual(stats_dict_user1['Negative Messages'], 1, "Expected at least 1 negative message for User1")

        # Unique words count for User1:
        # "good", "morning", "feeling", "happy", "today", "let's", "make", "this", "a", "great", "day",
        # "oh", "no", "what", "happened", "bad", "news", "is", "not" (stopwords are removed by CountVectorizer in basic_stats)
        # This needs careful checking against how basic_stats calculates unique_words_count (stopwords list etc.)
        # For now, just check it's a plausible number (positive integer)
        self.assertIsInstance(stats_dict_user1['Unique Words Count'], int)
        self.assertGreater(stats_dict_user1['Unique Words Count'], 5) # Conservative check


    def test_basic_stats_formatted_strings(self):
        chat_lines = [
            "10/10/22, 09:00 AM - User1: test one two",
            "10/10/22, 09:05 AM - User1: test three"
        ]
        clean_df = self._create_and_parse_chat_file(chat_lines)
        stats_dict_user1 = basic_stats(clean_df.copy(), username='User1')

        self.assertIsInstance(stats_dict_user1['Common Unigrams'], str)
        self.assertTrue("<li>test: 2</li>" in stats_dict_user1['Common Unigrams'] or "<li>test: 2</li>" in stats_dict_user1['Common Unigrams'])
        self.assertIsInstance(stats_dict_user1['Hindi Abuse Counts'], str)
        self.assertEqual(stats_dict_user1['Hindi Abuse Counts'], "<ul></ul>") # Assuming no abuse words

    def test_basic_stats_runs_without_error(self):
        """ Test that basic_stats completes without error for a typical case."""
        chat_lines = [
            "10/10/22, 10:00 AM - User1: Hello world http://example.com 😊",
            "10/10/22, 10:05 AM - User1: Second message",
            "10/11/22, 11:30 AM - User2: <Media omitted>",
            "10/12/22, 12:20 PM - User3: Test message with emoji 😱 and url www.test.com",
            "10/13/22, 01:25 PM - User1: <This message was edited>",
            "10/14/22, 02:30 PM - User2: This message was deleted"
        ]
        clean_df = self._create_and_parse_chat_file(chat_lines)
        
        try:
            # Test for a specific user
            if "User1" in clean_df['name'].unique():
                 basic_stats(clean_df.copy(), username='User1')
            # Test for all users (username=None)
            basic_stats(clean_df.copy()) 
        except Exception as e:
            self.fail(f"basic_stats raised an exception unexpectedly: {e}")

if __name__ == '__main__':
    unittest.main()
